#!/usr/bin/python
import toml
import argparse
import logging
import shutil
import sys
from os import path, environ
from hashlib import sha256
from collections import OrderedDict
from subprocess import Popen, PIPE

log = logging.getLogger(__name__)


def exhandler(function, parser):
    """If -examples was specified in 'args', the specified function
    is called and the application exits.

    :arg function: the function that prints the examples.
    :arg parser: the initialized instance of the parser that has the
      additional, script-specific parameters.
    """
    args = vars(bparser.parse_known_args()[0])
    if args["examples"]:
        function()
        exit(0)

    args.update(vars(parser.parse_known_args()[0]))
    return args

def _common_parser():
    """Returns a parser with common command-line options for all the scripts
    in the IDT suite.
    """
    import argparse
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument("--examples", action="store_true",
                        help="See detailed help and examples for this script.")
    parser.add_argument("--verbose", action="store_true",
                        help="See verbose output as the script runs.")
    parser.add_argument("--debug", action="store_true",
                        help="Print verbose calculation information for debugging.")

    return parser


def examples():
    """Prints examples of using the script to the console using colored output.
    """
    print("ClearCOIN Validator Configuration Upgrade\n")
    print("upgrade_validators.py /etc/stellar/stellar-config.cfg\n")


bparser = _common_parser()
script_options = {
    "config": {"help": "Path to the file to update configuration for."},
    "--stub": {"help": "File to grab desired validator configuration from. "
               "If not specified, then validators will not be updated."},
    "--domain": {"help": "Edit the home domain settings for this validator. "
                 "If not specified, then home domain will not be updated.",
                 "action": "store_true"},
    "--output": {"help": "Path to write the modified file to; defaults to `config`."},
    "--backup": {"help": "Path to save a backup of the original file to just in case."}
}
"""dict: default command-line arguments and their
    :meth:`argparse.ArgumentParser.add_argument` keyword arguments.
"""


def _parser_options():
    """Parses the options and arguments from the command line."""
    pdescr = "ClearCOIN Configuration Updater"
    parser = argparse.ArgumentParser(parents=[bparser], description=pdescr)
    for arg, options in script_options.items():
        parser.add_argument(arg, **options)

    args = exhandler(examples, parser)
    if args is None:
        return

    if args["debug"]:
        logging.basicConfig(level=logging.DEBUG)

    return args


def _get_validators_list(args):
    """Grabs the list of validators that the config file should be 
    updated to include.
    """
    with open(args["stub"]) as f:
        config = toml.load(f)

    return config


def _get_existing_config(args):
    """Retrieves the existing validator configuration file that needs
    to be updated.
    """
    with open(args["config"]) as f:
        config = toml.load(f)

    return config


def _get_dict_hash(d):
    """Creates a deterministic hash of the specified dictionary for 
    equality comparison.
    """
    o = OrderedDict(d)
    s = repr(o)
    h = sha256()
    h.update(s.encode())
    return h.hexdigest()


def _get_dict_set(config, key):
    """Retrieves a set of hashes representing the objects at `key` in a given
    configuration file.
    """
    result, index = set(), {}
    for hd in config[key]:
        h = _get_dict_hash(hd)
        result.add(h)
        index[h] = hd

    return result, index

def _get_home_domain_set(config):
    """Retrieves a set of hashes representing the home domains in a given
    configuration file.
    """
    return _get_dict_set(config, "HOME_DOMAINS")


def _get_validator_set(config):
    """Retrieves a set of hashes representing the validators in a given
    configuration file.
    """
    return _get_dict_set(config, "VALIDATORS")


def _log_set_changes(d, c, label):
    """Logs the changes in the specified sets of configs.
    
    Args:
        d (tuple): of `(set, index)` for the *desired* configuration.
        c (tuple): of `(set, index)` for the *current* configuration.
        label (str): that these sets correspond to.
    """
    add, rm = d[0] - c[0], c[0] - d[0]
    n = len(add) + len(rm)
    log.info("{} Changes to {} Configuration".format(n, label))

    if len(add) > 0:
        log.debug("The following {} will be added:".format(label))
        for h in add:
            account = d[1][h]["PUBLIC_KEY"]
            domain = d[1][h]["HOME_DOMAIN"]
            log.debug("  {}@{}".format(account, domain))

    if len(rm) > 0:
        log.debug("The following {} will be removed:".format(label))
        for h in rm:
            account = c[1][h]["PUBLIC_KEY"]
            domain = c[1][h]["HOME_DOMAIN"]
            log.debug("  {}@{}".format(account, domain))


def _overwrite_config_file(args, desired, current):
    """Overwrites the entries in the `current` configuration using the
    `desired` one.
    """
    current["HOME_DOMAINS"] = desired["HOME_DOMAINS"]
    current["VALIDATORS"] = desired["VALIDATORS"]

    log.info("Writing modified configuration to {}.".format(args['output']))
    with open(args["output"], 'w') as f:
        toml.dump(current, f)



def execute(args, folder=None, wait=True, nlines=100, venv=None,
            printerr=True, env_vars=None, errignore=None, shell=True, **kwargs):
    """Executes the specified tuple that should include the command as
    first item and additional arguments afterward. See the
    documentation for :class:`subprocess.Popen` for details.
    Args:
        args (list): of `str`; first item should be command to
          execute; additional arguments following.
        folder (str): directory to switch into before executing the
          command.
        wait (bool): when True, block the current thread until
          execution completes; otherwise, returns immediately.
        nlines (int): by default, `stdout` and `stderr` are redirected to
          :data:`subprocess.PIPE`. This is the maximum number of lines that will
          be returned for large outputs (so that memory doesn't get overwhelmed
          by large outputs).
        venv (str): when not `None`, the name of a virtualenv to
          activate before running the command.
        printerr (bool): when True, if `stderr` is not empty, print
          the lines automatically.
        env_vars (dict): of environment variables to set before calling the
          execution. The variables will be set back after execution.
        errignore (str): if errors are produced that include this pattern, then
          they will *not* be printed to `stdout`.
        kwargs (dict): additional arguments that are passed directly
          to the :class:`subprocess.Popen` constructor.
    Returns:
        dict: with keys ['process', 'stdout', 'stderr'], where 'process' is the
        instance of the subprocess that was created; 'stdout' and 'stderr' are
        only included if they were set to :data:`subprocess.PIPE`.
    .. note:: If the output from 'stdout' and 'stderr' are too large, only the
      first 100 lines will be returned. Use parameter `nlines` to control output
      size.
    """
    # from subprocess import Popen, PIPE
    if "stdout" not in kwargs:
        kwargs["stdout"] = PIPE
    if "stderr" not in kwargs:
        kwargs["stderr"] = PIPE
    if folder is not None:
        kwargs["cwd"] = folder

    if venv is not None: # pragma: no cover No guarantee that virtual
                         # envs exist on testing machine.
        prefix = path.dirname(sys.executable)
        if isinstance(venv, str):
            base_home = environ.get("WORKON_HOME")
            if base_home is not None:
                venv_home = path.abspath(path.expanduser(base_home))
                prefix = path.join(venv_home, venv, "bin")
        args[0] = path.join(prefix, args[0])

    oldvars = {}
    if env_vars is not None:
        for name, val in env_vars.items():
            oldvars[name] = environ[name] if name in environ else None
            environ[name] = val

    log.info("Executing `{}` in {}.".format(' '.join(args), folder), 2)
    pexec = Popen(' '.join(args), shell=shell, executable="/bin/bash", **kwargs)

    if wait:
        sout, serr = pexec.communicate()
    else:
        sout, serr = b"", b""

    if env_vars is not None:
        #Set the environment variables back to what they used to be.
        for name, val in oldvars.items():
            if val is None:
                del environ[name]
            else:
                environ[name] = val

    #Redirect the output and errors so that we don't pollute stdout.
    output = None
    if kwargs["stdout"] is PIPE:
        output = []
        for line in sout.decode("ascii").split('\n'):
            #Filter non fatal exceptions such as future warnings. A full list can be found here
            # https://docs.python.org/3/library/exepctions.html#exception-hierarchy
            if line.strip() == "":
                continue

            if not ("FutureWarning" in line or "import" in line or "\x1b[0m" in line):
                output.append(line)
                if len(output) >= nlines:
                    break

    error = None
    if kwargs["stderr"] is PIPE and serr != '':
        error = []
        for line in serr.strip().decode("ascii").split('\n'):
            if (errignore is None or errignore not in line) and line.strip() != "":
                error.append(line)
            if len(error) >= nlines:
                break

        if printerr and len(error) > 0 and all([isinstance(i, str) for i in error]):
            log.error(''.join(error))

    return {
        "process": pexec,
        "output": output,
        "error": error
    }


def get_internet_hostname():
    """Gets the configured dynamic DNS address of the host from
    `/etc/clearos/network.conf
    """
    args = ["cat", "/etc/clearos/network.conf"]
    out = execute(args)
    result = [l.strip() for l in out["output"] if "INTERNET_HOSTNAME" in l]
    if len(result) == 1 and '=' in result[0]:
        return result[0].split('=')[1].replace('"', '')
    else:
        log.warning("Couldn't find the internet hostname for DDNS.")


def _update_home_domain(args, config):
    """Updates the home domain for the current node using the dynamic DNS address
    from  /etc/clearos/network.conf.
    """
    domain = get_internet_hostname()
    existing = config["NODE_HOME_DOMAIN"]

    if domain is not None:
        log.debug("Overwriting NODE_HOME_DOMAIN with {}".format(domain))
        config["NODE_HOME_DOMAIN"] = domain

        hc = _get_home_domain_set(config)
        vc = _get_validator_set(config)

        for home in hc:
            if home["HOME_DOMAIN"] == existing:
                log.debug("Overwriting home domain in {} with {}.".format(home, domain))
                home["HOME_DOMAIN"] = domain

        for val in vc:
            if val["HOME_DOMAIN"] == existing:
                log.debug("Overwriting validator domain in {} with {}.".format(val, domain))
                val["HOME_DOMAIN"] = domain

        log.info("Writing modified configuration to {}.".format(args['output']))
        with open(args["output"], 'w') as f:
            toml.dump(config, f)



def run(**args):
    """Updates the configuration of the validators.
    """    
    current = _get_existing_config(args)

    if args["stub"]:
        desired = _get_validators_list(args)
        hd, hc = _get_home_domain_set(desired), _get_home_domain_set(current)
        vd, vc = _get_validator_set(desired), _get_validator_set(current)

        # Find the changes that we have to make so we can log them.
        _log_set_changes(hd, hc, "Home Domain")
        _log_set_changes(vd, vc, "Validators")

        _overwrite_config_file(args, desired, current)

    elif args["domain"]:
        _update_home_domain(args, current)


def main(args=None):
    args = _parser_options() if args is None else args
    args["config"] = path.abspath(path.expanduser(args["config"]))
    args["stub"] = path.abspath(path.expanduser(args["stub"]))
    if args["output"] is None:
        args["output"] = args["config"]

    if args["backup"] is not None:
        args["backup"] = path.abspath(path.expanduser(args["backup"]))
        shutil.copyfile(args["config"], args["backup"])

    return run(**args)


if __name__ == '__main__':  # pragma: no cover
    main()
